#!/usr/bin/env python

import os
import sys
import re
import subprocess
import shutil
import tempfile
import zipfile
import yaml
from .cloud_utils import execute
from .gcloud import hail_paths
from .slackconnector import SlackConnector

from IPython.core import magic_arguments
from IPython.core.magic import (Magics, magics_class, line_magic, cell_magic, line_cell_magic)
from IPython.utils.io import capture_output

import logging
logging.basicConfig(level = logging.INFO)
logger = logging.getLogger(__name__)


class HailRunner(object):
    def __init__(self, outdir="", input_file="", localdata="",
                 hail_jar_path="", pyhail_zip="",
                 header="", other_script_paths="", str_format='%'):
        
        # Assigns outdir and script_path
        self.outdir = os.path.dirname(outdir)
        
        # Define version
        self.version = 'debug' if 'VERSION' not in os.environ else os.environ['VERSION']

        # Create outdir and basic .jobs file structure
        if not os.path.isdir(self.outdir) or not os.path.isdir(os.path.join(self.outdir, 'log', self.version)):
            os.makedirs(os.path.join(self.outdir, "log", self.version) + '/')
        self.logdir = os.path.join(self.outdir, 'log', self.version)

        if not os.path.isdir(os.path.join(self.outdir, 'script', self.version)):
            os.makedirs(os.path.join(self.outdir, "script", self.version) + '/')
        self.scriptdir = os.path.join(self.outdir, 'script', self.version)

        # add current links
        if os.path.isdir(os.path.join(self.outdir, "current")):
            shutil.rmtree(os.path.join(self.outdir, "current"))
        os.makedirs(os.path.join(self.outdir, "current"))
        os.symlink(self.logdir, os.path.join(self.outdir, 'current', 'log'), target_is_directory=True)
        os.symlink(self.scriptdir, os.path.join(self.outdir, 'current', 'script'), target_is_directory=True)

        # add local /data folder links
        if localdata != '':
            if os.path.islink(os.path.join(self.outdir, 'data')):
                os.remove('data')
            os.symlink(localdata, os.path.join(self.outdir, 'data'), target_is_directory=True)

        # 0-based counter for submits made in script. Used for naming and tracking. 
        self.current_run = 0
        
        # 0-based counter that stores a particular run count for restoration later
        self._setr = []
        self._sig_popr = None
    
        # Initialize run-specific array attributes of PyRunner class
        self.current_job = []
    
        # Initialize Hail-related variables
        self.hail_jar_path = hail_jar_path
        self.hail_jar_base_name = os.path.basename(self.hail_jar_path)
        self.py_files = [ pyhail_zip ]
        if other_script_paths:
            if isinstance(other_script_paths, list):
                self.py_files.extend(other_script_paths)
            else:
                self.py_files.append(other_script_paths)
        self.py_files = ','.join(self.py_files)
        self.header = header
            
        # Read in inputfofn can be a list of files or read from a file 
        self.fofn = []
        if isinstance(input_file, str):
            self.fofn.append(input_file)
        else:
            self.fofn.append('')

        # Define string format rule
        self.str_format = '{' if str_format == '{' else '%'

        # set default as non-local mode
        self.local = False

        # init SlackConnector
        self.sc = SlackConnector('HailRunner')
        
    def chain_submit(self, name, command, outfile, *args):
        """ add_batch adds a batch of jobs to job_array in the form of method, output filename, input filename
        and method arguments. It maps 1-1 from previous fofn. All new output filenames are added to fofn[done_run+1] """
        # Initialize a new set of output fofn (if it does not yet exist)
        if len(self.fofn) == self.current_run + 1:
            self.fofn.append([])
        r = self.current_run if not self._sig_popr else self._setr[-1] # new code for use with setr
        self.current_job = [ command, outfile, self.fofn[r] ] + list(args)
        self.fofn[self.current_run + 1] = outfile
        self.submit(name)
        
    def submit_job(self, name, command, outfile, *args):
        # Initialize a new set of output fofn (if it does not yet exist)
        if len(self.fofn) == self.current_run + 1:
            self.fofn.append([])
        self.current_job = [ command, outfile ] + list(args)
        self.fofn[self.current_run + 1] = outfile 
        self.submit(name)
        
    def file_exists(self, path):
        return(os.path.exists(path))
    
    def submit(self, name):
        # Check if job completed
        if self.file_exists(self.current_job[1]):
            logger.info('Run {} successfully completed in an earlier attempt.'.format(self.current_run + 1))
            self.done()
        # Check if this step is marked "done" in log folder
        elif os.path.isfile(os.path.join(self.logdir, 'run{0}.done'.format(self.current_run + 1))):
            logger.info('Skipping checks for run {0}...'.format(self.current_run + 1))     
        # Submit Spark command
        else:
            # Specify script path
            script_path = os.path.join(
                self.scriptdir,
                "run{}.{}.py".format(self.current_run + 1, name)
            )

            # Copy Python script over if it is a file
            if os.path.isfile(self.current_job[0]):
                shutil.copy2(self.current_job[0], script_path)
            # Write command string as Python script
            else:
                with open(script_path, 'w') as f:
                    f.write(self.header)
                    input_list = self.current_job[2:] + [ self.current_job[1] ]
                    if self.str_format == '%':
                        f.write(self.current_job[0] % tuple(input_list))
                    else:
                        f.write(self.current_job[0].format(*input_list))
            
            # Specify log file
            log_path = os.path.join(self.logdir, "run{}.{}.o".format(self.current_run + 1, name))
            
            try:
                self.spark_submit(script_path, log_path)
            except Exception as e:
                self.sc.notify("run{}.{} has failed.".format(self.current_run + 1, name))
                sys.stderr.write(e)
                sys.exit(1)
            
            self.sc.notify("run{}.{} has finished successfully.".format(self.current_run + 1, name))
            self.done()

        self.current_run += 1
        if self._sig_popr:
            self._sig_popr = False
            self._setr.pop()

        # reset to non-local mode
        self.local = False
    
    def add_pyfile(self, path):
        self.py_files = self.py_files + ",%s" % path

    def quick_submit(self, command):
        temp_py = tempfile.mkstemp(suffix='.py')
        if not os.path.isfile(command):
            with open(temp_py[1], 'w') as f:
                f.write(self.header)
                f.write(command)
        else:
            shutil.copy2(command, temp_py[1])
        self.spark_submit(temp_py[1], False)
        if temp_py is not None:
            os.remove(temp_py[1])
        self.sc.notify("Quick submit has finished successfully.")
        # reset to non-local mode
        self.local = False
            
    # def spark_submit(self, script_path, log_path):
    #     py_files = process_pyfiles(py_files)
    #     cmd = [ 'spark-submit', script_path,
    #             '--files={}'.format(self.hail_jar_path),
    #             '--py-files={}'.format(self.py_files),
    #             '--properties=spark.driver.extraClassPath=./{0},spark.executor.extraClassPath=./{0}'.format(self.hail_jar_base_name),
    #             '--driver-log-levels', 'root=FATAL,is.hail=INFO' ]
    #     execute(cmd, logger = log_path, clean_spark_log = True)
    
    def done(self, abspath = False):
        with open(os.path.join(self.logdir, 'output.fofn'), 'w') as f:
            if abspath:
                fofn = [ os.path.abspath(filename) for filename in self.fofn ]
            else:
                fofn = self.fofn
            f.write('\n'.join(fofn)+'\n')

    def exit(self):
        sys.exit('Exiting script')

    def set_skip(self):
        f = open(os.path.join(self.outdir, 'run{0}.done'.format(self.current_run)), 'w')
        f.write('skip...')
        f.close()

    def reset(self):
        path = self.f
        runner.fofn = self.fofn[0:-1]
        self.current_run -= 1
        return(path)

    @property
    def r(self):
        return(self.current_run)

    @property
    def f(self):
        if len(self.fofn[self.current_run]) == 1:
            return(self.fofn[self.current_run][0])
        return(self.fofn[self.current_run])
    
    @property 
    def popr(self):
        self._sig_popr = True
        return(self._setr[-1])
    
    @property
    def popf(self):
        self._sig_popr = True
        if len(self.fofn[self._setr[-1]]) == 1:
            return(self.fofn[self._setr[-1]][0])
        return(self.fofn[self._setr[-1]])
    
    # should be setter, but it sort of behaves differently
    def pushr(self):
        self._setr.append(self.current_run)

    @property
    def past(self):
        paths = [ path.strip() for path in execute(['cat', os.path.join(self.logdir, 'output.fofn')], 
            return_out = True, silent = True) ]
        return(list(enumerate(paths)))

class HailRunnerGC(HailRunner):
    def __init__(self, instance, header=None, default_reference='GRCh37', other_files=None, tracker=False, str_format=None, version=2, **args):
        hail_jar, pyhail_zip = hail_paths(version = version)    
        header = header if header else get_hail_header(default_reference = default_reference)
        str_format = '{' if str_format == None and version == 1 else '%'
        super(HailRunnerGC, self).__init__(hail_jar_path = hail_jar,
                                           pyhail_zip = pyhail_zip,
                                           header = header, str_format = str_format,
                                           **args)
        self.instance = instance 

        if isinstance(other_files, list):
            self.other_files = other_files.join(',') 
        elif isinstance(other_files, str):
            self.other_files = other_files
        else:
            self.other_files = None

        if tracker and os.environ['TRACKER_PATH']:
            if self.other_files:
                self.other_files = f"{self.other_files},{os.environ['TRACKER_PATH']}"
            else:
                self.other_files = f"{os.environ['TRACKER_PATH']}"
            # tracker/db.json as copied to dataproc is in a special nested directory, but can be referred to directly as the current working directory
            self.header += "\nos.environ['TRACKER_PATH'] = 'db.json'\n\n"

    def file_exists(self, path):
        try:
            if re.search(r'\.ht$|\.mt$', path):
                path = os.path.join(path, '_SUCCESS')
            if execute(['gsutil', 'ls', path], return_out=True, silent=True):
                return(True)
        except:
            return(False)

    def spark_submit(self, script_path, log_path, skip_jar = True, add_py_file = None):
        cmd = [ 'gcloud', 'dataproc', 'jobs', 'submit', 'pyspark', script_path, '--cluster', self.instance ]
        
        # Add additional script files to load into working directory
        py_files = self.py_files.split(",")
        if add_py_file:
            if isinstance(add_py_file, list):
                py_files.extend(add_py_file)
            else:
                py_files.append(add_py_file)
        py_files = process_pyfiles(py_files)
        if skip_jar and len(py_files) > 1:
            cmd.extend([ '--py-files={}'.format(",".join(py_files[1:])) ])
        else:
            cmd.extend([ '--py-files={}'.format(",".join(py_files))
                #'--files={}'.format(self.hail_jar_path),
                #'--properties=spark.driver.extraClassPath=./{0},spark.executor.extraClassPath=./{0}'.format(self.hail_jar_base_name),
                #'--p', 'root=FATAL,is.hail=INFO' 
            ])
        if self.other_files:
            cmd.extend([ '--files', self.other_files ])
        # set to local mode here
        cmd = cmd if not self.local else [ 'python',  script_path ]
        execute(cmd, logger = log_path, clean_spark_log = True)
    
    def copy_log(self, target=None):
        if not target:
            target = self.logdir
        cmd = [ 'gcloud', 'compute', 'scp', f'{self.instance}-m:/home/hail/hailrunner.log', target ] #, '--zone', zone ]
        execute(cmd, silent=False)

def process_pyfiles(py_files):
    '''
    py_files is a list of Python files and zipped files.
    process_pyfiles zips all the Python files into a .zip file 
    that will be unpackaged in the PYTHONPATH of the dataproc cluster.
    '''
    return_list = []    
    temp_zip = tempfile.mkstemp(suffix = '.zip', prefix = 'pyfile_')[1]
    zf = zipfile.ZipFile(temp_zip, 'w', zipfile.ZIP_DEFLATED)
    for path in py_files:
        if path.endswith('.zip') or path.endswith('.egg'):
            return_list.append(path)
        if path.endswith('.py'):
            zf.write(path, arcname = os.path.basename(path))
        else:
            for root, _, pyfiles_walk in os.walk(path):
                for pyfile in pyfiles_walk:
                    if pyfile.endswith('.py'):
                        zf.write(os.path.join(root, pyfile),
                            os.path.relpath(os.path.join(root, pyfile), os.path.join(path, '..')))
    return_list.append(temp_zip)
    return(return_list)

def get_hail_header(log_file=None,
                    default_reference='GRCh37',
                    n_cpu=None,
                    requester_pays=False,
                    tmp_dir=None,
                    import_lines=None,
                    version=None):
    """
    Return the Hail header for a given configuration.

    Parameters
    ----------
    log_file : str or None, optional
        The path to the log file. If None, a default log file is used.
    default_reference : str, optional
        The default reference genome to use. Default is 'GRCh37'.
    n_cpu : int or None, optional
        The number of CPUs to use. If None, all available CPUs are used.
    requester_pays : bool, optional
        Whether to use requester-pays mode for Google Cloud Storage. Default is 'False'.
    tmp_dir : str or None, optional
        The path to the temporary directory. If None, a default directory is used.
    import_lines : list or None, optional
        A list of extra import statements to include in the header.
    version : str or None, optional
        The version of Hail to use. If None, the latest version is used.

    Returns
    -------
    str
        The Hail header as a string.

    Raises
    ------
    ValueError
        If an invalid configuration is provided.

    Examples
    --------
    >>> from hailrunner import get_hail_header
    >>> header = get_hail_header(default_reference='GRCh38',
    ...                          n_cpu=4,
    ...                          requester_pays=False,
    ...                          tmp_dir='gs://tsingh-tmp/my_tmp_dir',
    ...                          import_lines=['import numpy as np', 'import pandas as pd'])
    """
    config_str = [ 'min_block_size=0', f"default_reference='{default_reference}'" ]
    if log_file is not None:
        config_str.append(f"log='{log_file}'")
    elif n_cpu is None:
        config_str.append("log='/home/hail/hailrunner.log'")
    elif n_cpu is not None:
        config_str.append("log='hailrunner.log'")
    if requester_pays:
        config_str.append("spark_conf={'spark.hadoop.fs.gs.requester.pays.mode': 'AUTO'}")
    if tmp_dir:
        config_str.append(f"tmp_dir='{tmp_dir}'")
    if n_cpu is not None:
        config_str.append(f"master='local[{n_cpu}]'")
    config_str = ', '.join(config_str)
    header_str = """
import os
import re
import sys

import hail as hl
hl.init(%s)
from pprint import pprint

import logging
logging.basicConfig(level = logging.INFO)
logger = logging.getLogger(__name__)

logger.info('Default reference: {}'.format(hl.default_reference().name))

""" % config_str

    if import_lines is not None:
        for line in import_lines:
            header_str += f'{line}\n'

    return header_str

def get_hail_utils_header():
    return("""
import os
import re
import sys

import hail as hl
from pprint import pprint

import logging
logging.basicConfig(level = logging.INFO)
logger = logging.getLogger(__name__)
""")

def write_hail_utils_header(utils_file, lines_str):
    with open(utils_file, 'w') as f:
        f.write(get_hail_utils_header() + "\n" + lines_str + "\n")

# from IPython.display import Javascript
# Javascript(cwnb) os.path.abspath(notebookName) + './log'

cwnb = """
var kernel = Jupyter.notebook.kernel; 
var command = ["notebookPath = ",
               "'", window.document.body.dataset.notebookPath, "'" ].join('')
//alert(command)
kernel.execute(command)
var command = ["notebookName = ",
               "'", window.document.body.dataset.notebookName, "'" ].join('')
//alert(command)
kernel.execute(command)
"""

# http://mlexplained.com/2017/12/28/creating-custom-magic-commands-in-jupyter/
# https: // github.com/ipython/ipython/issues/11826
@magics_class
class RunnerMagics(Magics):  
    def __init__(self, shell, runner):
        super(RunnerMagics, self).__init__(shell)
        self.runner = runner
        
    @cell_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument('--name', '-n', help='Name.')
    @magic_arguments.argument('--out', '-o', help='Outfile.')
    @magic_arguments.argument('--args', '-a', help='Infile and other arguments, comma-separated.')
    @magic_arguments.argument('--silent', '-q', action = 'store_true', default = False, help = 'Hide results in variable captured.')
    @magic_arguments.argument('--local', '-l', action = 'store_true', default = False, help = 'Run code in local mode using Python.')
    @magic_arguments.argument('--fstring', '-f', action='store_true', default=False, help='Interpret f-strings in code.')
    def submit(self, line, cell):
        args = magic_arguments.parse_argstring(self.submit, line)
        if args.local:
            self.runner.local = True
        if args.fstring:
            cell = cell.format(**self.shell.user_ns)
        if args.silent:
            with capture_output() as runlog:
                if args.args:
                    self.runner.submit_job(args.name, cell, args.out, *args.args.split(","))
                else:
                    self.runner.submit_job(args.name, cell, args.out)
                self.shell.user_ns['runlog'] = runlog
        else:
            if args.args:
                self.runner.submit_job(args.name, cell, args.out, *args.args.split(","))
            else:
                self.runner.submit_job(args.name, cell, args.out)      

    @cell_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument('--name', '-n', help='Name.')
    @magic_arguments.argument('--out', '-o', help='Outfile.')
    @magic_arguments.argument('--args', '-a', help='Infile and other arguments, comma-separated.')
    @magic_arguments.argument('--silent', '-q', action = 'store_true', default = False, help = 'Hide results in variable captured.')
    @magic_arguments.argument('--local', '-l', action = 'store_true', default = False, help = 'Run code in local mode using Python.')
    @magic_arguments.argument('--fstring', '-f', action='store_true', default=False, help='Interpret f-strings in code.')
    def chain(self, line, cell):
        args = magic_arguments.parse_argstring(self.chain, line)
        if args.local:
            self.runner.local = True
        if args.fstring:
            cell = cell.format(**self.shell.user_ns)
        if args.silent:
            with capture_output() as runlog:
                if args.args:
                    self.runner.chain_submit(args.name, cell, args.out, *args.args.split(","))
                else:
                    self.runner.chain_submit(args.name, cell, args.out)
                self.shell.user_ns['runlog'] = runlog
        else:
            if args.args:
                self.runner.chain_submit(args.name, cell, args.out, *args.args.split(","))
            else:
                self.runner.chain_submit(args.name, cell, args.out)         
    
    @cell_magic
    @magic_arguments.magic_arguments()
    @magic_arguments.argument('--args', '-a', help='Infile and other arguments, comma-separated.')
    @magic_arguments.argument('--silent', '-q', action = 'store_true', default = False, help = 'Hide results in variable captured.')
    @magic_arguments.argument('--local', '-l', action = 'store_true', default = False, help = 'Run code in local mode using Python.')
    @magic_arguments.argument('--fstring', '-f', action='store_true', default=False, help='Interpret f-strings in code.')
    def quick(self, line, cell):
        args = magic_arguments.parse_argstring(self.quick, line)
        if args.local:
            logger.info('Run as local mode.')
            self.runner.local = True
        if args.fstring:
            cell = cell.format(**self.shell.user_ns)
        if args.silent:
            with capture_output() as runlog:
                if args.args:
                    self.runner.quick_submit(cell % tuple(args.args.split(",")))
                else:
                    self.runner.quick_submit(cell)
            self.shell.user_ns['runlog'] = runlog
        else:
            if args.args:
                self.runner.quick_submit(cell % tuple(args.args.split(",")))
            else:
                self.runner.quick_submit(cell)



#!/usr/bin/env python

import os
import re
import sys
import subprocess

os.environ["PYTHONUNBUFFERED"] = "1"

def execute(cmd, logger = None, return_out = False, silent = False, 
            return_cmd = False,
            shell = False, clean_spark_log = False):
    if not shell:
        cmd = [ str(c) for c in cmd ]

    if return_cmd:
        return ' '.join(cmd)

    process = subprocess.Popen(cmd,
                               stdout = subprocess.PIPE,
                               stderr = subprocess.STDOUT,
                               bufsize = 0, #1, for buffered
                               universal_newlines = True,
                               shell = shell)
   
    f = open(logger, 'w') if logger else None

    output = []
    clean_flag = 0
    for line in iter(process.stdout.readline, ''):   
        #----
        # writing to screen   
        if not silent:
            sys.stdout.write(line)
            sys.stdout.flush()
    
        #----
        # writing to log
        if f and not clean_spark_log:
            # write everything
            f.write(line)
            f.flush()
        elif f and clean_spark_log:
            if re.search('^Welcome to', line):
                # trigger the start of the Hail run
                clean_flag = 1
                f.write(line)
            elif clean_flag == 1:
                # in the Hail run
                if line == '\n' or line == '' or re.search('^\s+$', line):
                    continue
                elif line[0:7] != '[Stage ':
                    # write non-empty line
                    f.write(line)
                elif line[0:7] == '[Stage ':
                    # Spark stage line, clean up, and write if not empty
                    l = re.sub('\[Stage\s.*\]', '', line)
                    if l != '\n' and l != '' and not re.search('\s+', l):
                        f.write(l)
            else:
                if re.search('^spark\.', line) or re.search('SparkContext', line):
                    f.write(line)
            if re.search('Job.*finished successfully.', line):
                break
            f.flush()
        if return_out:
            output.append(line)

    process.stdout.close()
    
    return_code = process.wait()
    if return_code:
        if silent and return_out:
            sys.stderr.write(''.join(output))
        raise subprocess.CalledProcessError(return_code, cmd)
       
    if return_out:
        return output



#!/usr/bin/python

from pycloud.cloud_utils import execute

def execute_hdfs_cmd(options):
    cmd = [ 'hadoop', 'fs' ] + options
    execute(cmd)

def ls(path):
    execute_hdfs_cmd(['-ls', '-h', path])

def lsr(path):
    execute_hdfs_cmd(['-ls', '-h', '-R', path])

def du(path):
    execute_hdfs_cmd([ '-du', '-s', '-h', path ])
    
def cp(from_path, to_path):
    execute_hdfs_cmd(['-cp', from_path, to_path])

def mv(from_path, to_path):
    execute_hdfs_cmd(['-mv', from_path, to_path])

    
def chmod(path, code = '700', recursive = False):
    options = [ '-chmod', code ]
    if recursive:
        options.append('-r')
    execute_hdfs_cmd(options + path)

def rm(path, recursive = False, force = False, quiet = False):
    options = [ '-rm' ]
    if recursive:
        options.append('-r')
    if force:
        options.append('-f')
    execute_hdfs_cmd(options + [ path ])
        
def rmdir(path):
    execute_hdfs_cmd([ '-rmdir', path ])        

def mkdir(path):
    execute_hdfs_cmd([ '-mkdir', '-p', path ])
    
def get(remote_path, local_path):
    """Retrieve files from hdfs/Spark file system.
    
    Specify the file or directory on the hdfs and the path to the local directory
    to copy it to.
    
    Parameters
    ----------
    remote_path : {str}
        Should start with `/user/` for the Broad system
    local_path : {str}
        Should start with file:///psych/genetics_data/tsingh/projects/sczexomes/analysis/ for the Broad system
    """
    execute_hdfs_cmd(['-get', remote_path, local_path])

def put(local_path, remote_path):
    """Copy files to the hdfs/Spark file system.
    
    Specify the file or directory on the local system to the hdfs.
    
    Parameters
    ----------
    local_path : {str}
        Should start with file:///psych/genetics_data/tsingh/projects/sczexomes/analysis/ for the Broad system
    remote_path : {str}
        Should start with `/user/` for the Broad system
    """
    execute_hdfs_cmd(['-put', local_path, remote_path])
    
def put_directory(local_dir, remote_dir):
    execute_hdfs_cmd([ '-copyFromLocal', local_dir, remote_dir ])

def get_directory(remote_dir, local_dir):
    execute_hdfs_cmd([ '-copyToLocal', remote_dir, local_dir ])

def glob(path):
    lines = execute([ 'hadoop', 'fs', '-ls', path], return_out = True, silent = True)
    return([ line.split()[-1] for line in lines[1:] ])



#!/usr/bin/env python

import os
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from loguru import logger

class SlackConnector():
    """A class for connecting to Slack and sending notifications.

    Parameters
    ----------
    name : str, optional
        The name of the SlackConnector instance. Defaults to 'SlackConnector'.

    Attributes
    ----------
    sc : slack_sdk.web.client.WebClient
        The Slack web client used for sending notifications.
    slack_user_id : str
        The ID of the Slack user to send notifications to.
    name : str
        The name of the SlackConnector instance.

    Raises
    ------
    Exception
        If the Slack token is found but the connection fails.

    Examples
    --------
    >>> import slackconnector
    >>> sc = slackconnector.SlackConnector('name')
    >>> sc.notify('Hello world.')

    """
    def __init__(self, name = None):
        self.sc, self.slack_user_id = None, None
        self.name = name if name else 'Runner'
        if 'SLACK_BOT_TOKEN' in os.environ and 'SLACK_USER_ID' in os.environ:
            self.sc = WebClient(token=os.environ.get('SLACK_BOT_TOKEN'))
            self.slack_user_id = os.environ.get('SLACK_USER_ID')
            try:
                response = self.sc.auth_test()
                logger.info('Connected to Slack.')
            except SlackApiError as e:
                print(f"Error: {e.response['error']}")
                raise Exception('Slack token found but connection failed.')
        else:
            logger.info('SLACK_BOT_TOKEN and SLACK_USER_ID not found. Cannot connect to Slack.')
    
    def notify(self, msg):
        """Send a notification to a Slack user.

        Parameters
        ----------
        msg : str
            The message to send.

        Raises
        ------
        SlackApiError
            If the notification fails to send.

        Examples
        --------
        >>> sc.notify('Hello world.')

        """
        try:
            m = self.sc.chat_postMessage(channel=self.slack_user_id, text = msg, username=self.name) # , as_user = False 
            assert m["message"]["text"] == msg
        except SlackApiError as e:
            assert e.response["ok"] is False
            assert e.response["error"]
            logger.info(f"Got an error: {e.response['error']}")


#!/usr/bin/python

import os
import sys
import subprocess
import re
import tempfile
import logging
import shutil

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def execute(cmd, logger = None, return_out = False, silent = False, return_cmd = False, shell = False):
    if not shell:
        cmd = [ str(c) for c in cmd ]    

    if return_cmd:
        return(' '.join(cmd))

    process = subprocess.Popen(cmd, 
                               stdout = subprocess.PIPE, 
                               stderr = subprocess.STDOUT,
                               bufsize = 1,
                               universal_newlines = True,
                               shell = shell)
    
    f = open(logger, 'w') if logger else None
    
    output = []
    for line in iter(process.stdout.readline, ''):
        if not silent:
            sys.stdout.write(line)
            sys.stdout.flush()
        if f:
            f.write(line)
        if return_out:
            output.append(line)

    process.stdout.close()
    
    return_code = process.wait()
    if return_code:
        if silent and return_out:
            sys.stderr.write(''.join(output))
        raise subprocess.CalledProcessError(return_code, cmd)
        
    if return_out:
        return(output)

def unix_cmd(cmd, return_out = False, silent = False, shell = True,
             basic = True, huge_mem = False, print_error = False, # old options that aren't used
             strict = False, return_error = False):
    ''
    execute(cmd, return_out = return_out, silent = silent, shell = shell)

def archive(outdir, infile, code = 0o550):
    filename = os.path.basename(infile)
    outfile = os.path.join(outdir, filename)
    make_dir(outfile)
    shutil.copy2(infile, outdir)
    os.chmod(outfile, code)

def make_dir(outfile):
    """ make_dir checks if directory for outfile exists and creates it if necessary """
    outdir = os.path.dirname(os.path.abspath(outfile))
    try:
        os.makedirs(outdir)
    except OSError:
        if not os.path.isdir(outdir):
            raise
    return(outdir)

def rename(outfile):
    """ complete checks for existence of an outfile.part and renames it to outfile if it exists. """
    if os.path.isfile("{0}.part".format(outfile)):
        os.rename("{0}.part".format(outfile), outfile)
    else:
        raise Exception("Did not find {0}.part...".format(outfile))

def rename_ext(root, ext):
    os.rename("{0}.part.{1}".format(root, ext), "{0}.{1}".format(root, ext))

def add_tag(filelist, tag, dirname = None):
    if isinstance(filelist, str):
        return(__add_tag(filelist, tag, dirname))
    else:
        return([__add_tag(f, tag, dirname) for f in filelist])

def add_tag_chrom(filelist, tag, dirname, chrom_sep = ':'):
    if isinstance(filelist, str):
        filelist = [ filelist ]
    fofn = []
    for f in filelist:
        chrom = get_region(f, chrom = True, chrom_sep = chrom_sep)
        fofn.append(__add_tag(f, tag, os.path.join(dirname, chrom)))
    if len(fofn) == 1:
        fofn = fofn[0]
    return(fofn)       

def __add_tag(filepath, tag, dirname = None):
    filename = os.path.basename(filepath)
    if not dirname:
        dirname = os.path.dirname(filepath)
    filename_list = filename.split('.')
    filename_list.insert(1,tag)
    return(os.path.join(dirname, ".".join(filename_list)))

def change_ext(filelist, newext, oldext, dirname = None):
    if isinstance(filelist, str):
        return(__change_ext(filelist, newext, oldext, dirname))
    else:
        return([__change_ext(f, newext, oldext, dirname) for f in filelist])

def change_ext_chrom(filelist, newext, oldext, dirname, chrom_sep = ':'):
    if isinstance(filelist, str):
        filelist = [ filelist ]
    fofn = []
    for f in filelist:
        chrom = get_region(f, chrom = True, chrom_sep = chrom_sep)
        fofn.append(__change_ext(f, newext, oldext, os.path.join(dirname, chrom)))
    if len(fofn) == 1:
        fofn = fofn[0]
    return(fofn)       

def __change_ext(filepath, newext, oldext, dirname = None):
    filename = os.path.basename(filepath)
    if not dirname:
        dirname = os.path.dirname(filepath)
    filename_new = re.sub(r"{0}$".format(oldext), r"{0}".format(newext), filename)
    return(os.path.join(dirname, filename_new))

def add_ext(filelist, newext, dirname = None):
    if isinstance(filelist, str):
        return(__add_ext(filelist, newext, dirname))
    else:
        return([__add_ext(f, newext, dirname) for f in filelist])

def add_ext_chrom(filelist, newext, dirname, chrom_sep = ':'):
    if isinstance(filelist, str):
        filelist = [ filelist ]
    fofn = []
    for f in filelist:
        chrom = get_region(f, chrom = True, chrom_sep = chrom_sep)
        fofn.append(__add_ext(f, newext, os.path.join(dirname, chrom)))
    if len(fofn) == 1:
        fofn = fofn[0]
    return(fofn)       

def __add_ext(filepath, newext, dirname = None):
    filename = os.path.basename(filepath)
    if not dirname:
        dirname = os.path.dirname(filepath)
    return(os.path.join(dirname, filename+newext))

def change_outdir(filelist, dirname):
    if isinstance(filelist, str):
        return(__change_outdir(filelist, dirname))
    else:
        return([__change_outdir(f, dirname) for f in filelist])

def __change_outdir(filepath, dirname):
    filename = os.path.basename(filepath)
    return(os.path.join(dirname, filename))

def reset_tag(filelist, dirname = None, ext = None):
    if isinstance(filelist, str):
        return(__reset_tag(filelist, ext, dirname))
    else:
        return([__reset_tag(f, ext, dirname) for f in filelist])

def reset_tag_chrom(filelist, dirname, ext = None, chrom_sep = ':'):
    if isinstance(filelist, str):
        filelist = [ filelist ]
    fofn = []
    for f in filelist:
        chrom = get_region(f, chrom=True, chrom_sep = chrom_sep)
        fofn.append(__reset_tag(f, ext, os.path.join(dirname, chrom)))
    if len(fofn) == 1:
        fofn = fofn[0]
    return(fofn)       

def __reset_tag(filepath, ext = None, dirname = None, chrom_sep = ':'):
    """ ext must have a . """
    region = get_region(filepath, chrom_sep = chrom_sep)
    if not dirname:
        dirname = os.path.dirname(filepath)
    if not ext:
        ext = os.path.splitext(filepath)[1]
    return(os.path.join(dirname, region + ext))

def get_region(filepath, split = False, chrom = False, chrom_sep = ':'):
    filename = os.path.basename(filepath)
    regexp = re.compile("(?P<chrom>[0-9XY]{1,2})" + '\{0}'.format(chrom_sep) + "(?P<start>[0-9]+)\-(?P<end>[0-9]+)")
    match = regexp.search(filename)
    if match and split:
        return(match.group("chrom"), match.group("start"), match.group("end"))
    elif match and chrom:
        return(match.group("chrom"))
    else:
        filename_list = filename.split('.')
        return(filename_list[0])

def add_region(filelist, newregion, dirname):
    if isinstance(filelist, str):
        return(__add_region(filelist, newregion, dirname))
    else:
        return([__add_region(f, newregion, dirname) for f in filelist])

def __add_region(filepath, newregion, dirname = None):
    filename = os.path.basename(filepath)
    if not dirname:
        dirname = os.path.dirname(filepath)
    return(os.path.join(dirname, newregion + '.' + filename))

def replace_region(filepath, newregion, dirname = None):
    filename = os.path.basename(filepath)
    if not dirname:
        dirname = os.path.dirname(filepath)
    filename_lst = filename.split('.')
    filename_lst[0] = newregion
    return(os.path.join(dirname, '.'.join(filename_lst)))

def mktemp(outdir, suffix):
    return(outdir + tempfile.mktemp(suffix))

def decode_string(s):
    try:
        s.decode('ascii')
        return(s)
    except:
        logger.info('Replacing "%s" with "%s"', s, re.sub(r'[^\x00-\x7f]',r'', s))
        return(re.sub(r'[^\x00-\x7f]',r'', s))



#!/usr/bin/env python

import os
import sys
import re
import subprocess
import shutil
import tempfile
import yaml
from .cloud_utils import execute
from .slackconnector import SlackConnector

class GCloud():
    def __init__(self, account, project = None, local = True):
        self.account = account
        self.project = project
        if local:
            self.login()
            if project:
                self.set_project()
    def login(self):
        execute(['gcloud', 'auth', 'login', self.account])

    def logout(self):
        execute(['gcloud', 'auth', 'revoke'])
      
    def set_project(self):
        execute(['gcloud', 'config', 'set', 'project', self.project])
    
    def change_project(self, project):
        execute(['gcloud', 'config', 'set', 'project', self.project])
        self.project = project
    
    def ls(self, path, detailed = False, bucket = False):
        cmd = ['gsutil', 'ls']
        if detailed:
            cmd.append('-L')
        else:
            cmd.append('-l')
        if bucket:
            cmd.append('-b')
        cmd.append(path)
        execute(cmd)

    def cp(self, from_path, to_path, recursive = False, multithreaded = False, return_cmd = False, quiet = False):
        cmd = ['gsutil']
        if multithreaded:
            cmd.append('-m')
        if quiet:
            cmd.append('-q')
        cmd.append('cp')
        if recursive:
            cmd.append('-r')
        cmd.extend([from_path, to_path])
        if return_cmd:
            return(' '.join(cmd))
        else:
            execute(cmd)
    
    def cpi(self, paths, to_path, quiet = False):
        cmd = ['gsutil', '-m']
        if quiet:
            cmd.append('-q')
        cmd.append('cp')
        cmd.extend(paths)
        cmd.append(to_path)
        execute(cmd)
    
    def mv(self, from_path, to_path, keep_permissions = False, 
           multithreaded = False):
        cmd = ['gsutil', 'mv']
        if keep_permissions:
            cmd.append('-p')
        if multithreaded:
            cmd.append('-m')
        cmd.extend([from_path, to_path])
        execute(cmd)
        
    def du(self, path, total = True, human = True):
        cmd = ['gsutil', 'du', '-c']
        if human:
            cmd.append('-h')
        if total:
            cmd.append('-s')
        cmd.append(path)
        execute(cmd) 
    
    def create_bucket(self, path,
                      storage_class = 'regional', 
                      location = 'us-central1'):
        '''Storage classes: 
        https://cloud.google.com/storage/docs/storage-classes#regional
        '''
        cmd = ['gsutil', 'mb'] 
        cmd.append('-c {}'.format(storage_class))
        cmd.append('-l {}'.format(location))
        cmd.append('-p {}'.format(self.project))
        cmd.append(path)
        execute(cmd)
    
    def delete_bucket(self, path):
        '''Bucket must be empty first. Equivalent to rmdir'''
        execute(['gsutil', 'rb', path])
        
    def rm(self, path, recursive = False, force = False, quiet = False, multithreaded = False):
        '''* remove all objects in subdirectory. 
           ** removes all objects in subdirectory and those under them'''
        cmd = ['gsutil']
        if multithreaded:
            cmd.append('-m')
        cmd.append('rm')
        if recursive:
            cmd.append('-r')
        if force:
            cmd.append('-f')
        cmd.append(path)
        execute(cmd)
    
    def describe_project(self, path):
        execute(['gcloud', 'projects', 'get-iam-policy', self.project])
        execute(['gcloud', 'projects', 'describe', self.project])
    
    def list_projects(self):
        execute(['gcloud', 'projects', 'get-iam-policy', self.project])
    
    def rsync(self, from_path, to_path, 
              delete_too = False, recursive = False,
              multithreaded = False):
        cmd = ['gsutil', 'rsync'] 
        if delete_too:
            cmd.append('-d')
        if recursive:
            cmd.append('-r')
        if multithreaded:
            cmd.append('-m')
        cmd.extend([from_path, to_path])
        execute(cmd)
                
    def get_permission(self, path, recursive = False):
        cmd = ['gsutil', 'acl', 'get']
        if recursive:
            cmd.append('-r')
        cmd.append(path)
        execute(cmd)  
    
    def chmod(self, path, entity, 
              permission = None, 
              is_user = True, 
              is_group = False,
              is_project = False,
              add_modify = True, 
              remove = False,
              recursive = False, 
              multithreaded = False):
        ''' 
        Role must be: R, W, O.

        Sample uses:

        Everyone on the internet. 
        -g AllUsers:R 
        -g AllAuth:R

        Individual user:
        -u john.doe@example.com:W

        Group:
        -g admins@example.com:O
        -g broadinstitute.org:R # all Broad users

        Owners of example project:
        -p owners-example-project:W
        -p owners-12345:W # project number   
        '''
        cmd = ['gsutil']
        if multithreaded:
            cmd.append('-m')
        cmd.append('acl ch')
        if recursive:
            cmd.append('-r')
        if remove:
            cmd.extend(['-d', entity])
            cmd.append(path)
        else:
            if is_user:
                cmd.append('-u')
            elif is_group:
                cmd.append('-g')
            elif is_project:
                cmd.append('-p')
            cmd.append('{}:{}'.format(entity, permission))
        execute(cmd)
    
    def ComputeEngine(self):
        return(ComputeEngine(self.account, self.project))
    
class ComputeEngine():
    def __init__(self, account, project):
        self.account = account
        self.project = project
        self.instances = None
        self.get_instances(True)
        return
    
    def list_images(self):
        execute(['gcloud', 'compute', 'images', 'list'])
    
    def list_instances(self):
        execute(['gcloud', 'compute', 'instances', 'list'])
    
    def get_instances(self, silent = False):
        instances = {}
        output = execute(['gcloud', 'compute', 'instances', 'list'], return_out = True, silent = True)
        for instance in output[1:]:
            if instance == '':
                continue
            instance = instance.split()
            instances[instance[0]] = {'zone': instance[1], 'machine_type': instance[2], 'status': instance[-1] }
        self.instances = instances
        if not silent:
            print(instances)
    
    def ComputeEngineInstance(self, name):
        if name in self.instances:
            return(ComputeEngineInstance(instance = name,
                                         project = self.project,
                                         machine_type = self.instances[name]['machine_type'],
                                         zone = self.instances[name]['zone'], create = False))
        else:
            return(None)
                                                   
    def _list_machine_types(self):
        # g1-small: 1 cpu, 1.7 GB
        machine_types = ['f1-micro', 'g1-small']
        # n1-standard-x: x cpus, 3.75 GB per CPU
        machine_types.extend(['n1-standard-{}'.format(i) for i in [1, 2, 4, 8, 17, 32, 64]])
        # n1-highmem-x: x cpus, 6.5 GB per CPU
        machine_types.extend(['n1-highmem-{}'.format(i) for i in [2, 4, 8, 17, 32, 64]])
        # n1-highcpu-x: x cpus, 0.9 GB per CPU
        machine_types.extend(['n1-highcpu-{}'.format(i) for i in [2, 4, 8, 17, 32, 64]])
        self.machine_types = machine_types
    
    def show_machine_types(self):
        _list_machine_types()
        print(self.machine_types)
        
class ComputeEngineInstance():
    def __init__(self, instance, project,
                 zone = 'us-central1-a', 
                 machine_type = 'n1-highmem-4', 
                 create = True,
                 image = 'debian-8', # can be user image as well
                 boot_disk_size = None, 
                 description = None,
                 preemptible = False):
        '''
        https://cloud.google.com/compute/docs/images
        '''
        self.instance = instance
        self.project = project
        self.zone = zone
        self.machine_type = machine_type
        self.image = image
        self.boot_disk_size = None
        self.description = None
        self.preemptible = None
        if create:
            self.create_instance()
            
    def create_instance(self):
        cmd = ['gcloud', 'compute', 'instances', 'create']
        cmd.extend([self.instance])
        cmd.extend(['--zone', self.zone])
        cmd.extend(['--image {}'.format(self.image)])
        cmd.extend(['--machine-type', self.machine_type])
        if self.boot_disk_size:
            cmd.extend(['--boot-disk-size', self.boot_disk_size])
        if self.description:
            cmd.extend(['--description', self.description])
        if self.preemptible:
            cmd.extend(['--preemptible'])
        execute(cmd)
    
    def start(self):
        execute(['gcloud', 'compute', 'instances', 'start', self.instance, '--zone', self.zone])
        
    def stop(self):
        execute(['gcloud', 'compute', 'instances', 'stop', self.instance, '--zone', self.zone])
    
    def delete(self):
        execute(['gcloud', 'compute', 'instances', 'delete', self.instance, '--zone', self.zone])
        
    def reset(self):
        execute(['gcloud', 'compute', 'instances', 'reset', self.instance, '--zone', self.zone])
    
    def describe(self):
        output = execute(['gcloud', 'compute', 'instances', 'describe', 
                          self.instance, '--zone', self.zone],
                         silent = True, return_out = True)
        return(yaml.safe_load(''.join(output)))
    
    def ssh(self):
        '''Connecting securely
        https://cloud.google.com/solutions/connecting-securely'''
        cmd = ['gcloud', 'compute', '--project', self.project, self.zone, self.instance]
        print(' '.join(cmd))
        
    def tunnel(self, port = 2708, silent = False):
        cmd = ['gcloud', 'compute', 'ssh', self.instance]
        cmd.extend(['--project', self.project])
        cmd.extend(['--zone', self.zone])
        if silent:
            cmd.extend(['--ssh-flag=-f', '--ssh-flag=-N'])
        cmd.extend(['--ssh-flag=-L'])
        cmd.extend(['--ssh-flag={0}:localhost:{0}'.format(port)])
        print(' '.join(cmd))  
        
class DataProc():
    def __init__(self, account, project):
        self.account = account
        self.project = project
        self.instances = None
        self.get_instances(True)
        return
    
    def list_instances(self):
        execute(['gcloud', 'dataproc', 'clusters', 'list'])
    
    def get_instances(self, silent = False):
        instances = {}
        output = execute(['gcloud', 'dataproc', 'clusters', 'list'], return_out = True, silent = True)
        for instance in output[1:]:
            if instance == '':
                continue
            instance = instance.split()
            instances[instance[0]] = {'zone': instance[-1], 'worker_count': instance[1], 'status': instance[2] }
        self.instances = instances
        if not silent:
            print(instances)
    
    def describe_instance(self, instance):
        output = execute(['gcloud', 'dataproc', 'clusters', 'describe', instance], 
                         silent = True, return_out = True)
        return(yaml.safe_load(''.join(output)))
    
    def DataProcInstance(self, name):
        self.get_instances(True)
        if name in self.instances:
            desc = self.describe_instance(name)
            dpi = DataProcInstance(name,
                                    self.account,
                                    self.project,
                                    zone = desc['config']['gceClusterConfig']['zoneUri'].split('/')[-1],
                                    initialization_actions = ','.join([ ia['executableFile'] for ia in desc['config']['initializationActions'] ]),
                                    master_machine_type = desc['config']['masterConfig']['machineTypeUri'].split('/')[-1],
                                    master_boot_disk_size = desc['config']['masterConfig']['diskConfig']['bootDiskSizeGb'],
                                    n_workers = desc['config']['workerConfig']['numInstances'],
                                    worker_machine_type = desc['config']['workerConfig']['machineTypeUri'].split('/')[-1],
                                    n_worker_local_ssds = desc['config']['workerConfig']['diskConfig']['numLocalSsds'] if 'numLocalSsds' in desc['config']['workerConfig']['diskConfig'] else 0,
                                    worker_boot_disk_size = desc['config']['workerConfig']['diskConfig']['bootDiskSizeGb'],
                                    n_pre_workers = desc['config']['secondaryWorkerConfig']['numInstances'],
                                    #preworker_boot_disk_size = desc['config']['secondaryWorkerConfig']['diskConfig']['bootDiskSizeGb']
                                    version = desc['config']['softwareConfig']['imageVersion'],
                                    properties = desc['config']['softwareConfig']['properties'], 
                                  create = False)         
            return(dpi)
        else:
            return(None)
    
    def DataProcHailInstance(self, name):
        dpi = self.DataProcInstance(name)
        dpi.__class__ = DataProcHailInstance
        return(dpi)
    
    def _list_machine_types(self):
        # g1-small: 1 cpu, 1.7 GB
        machine_types = ['f1-micro', 'g1-small']
        # n1-standard-x: x cpus, 3.75 GB per CPU
        machine_types.extend(['n1-standard-{}'.format(i) for i in [1, 2, 4, 8, 17, 32, 64]])
        # n1-highmem-x: x cpus, 6.5 GB per CPU
        machine_types.extend(['n1-highmem-{}'.format(i) for i in [2, 4, 8, 17, 32, 64]])
        # n1-highcpu-x: x cpus, 0.9 GB per CPU
        machine_types.extend(['n1-highcpu-{}'.format(i) for i in [2, 4, 8, 17, 32, 64]])
        self.machine_types = machine_types
    
    def show_machine_types(self):
        _list_machine_types()
        print(self.machine_types)

def set_actions(hail = True, hail01 = False, jupyter = True, vep = False, vep_reference = 'GRCh37', repos = False, r = False):
    # /Users/tsingh/conda3/lib/python3.7/site-packages/hailctl/deploy.yaml
    actions = []
    if hail:
        actions.append('gs://hail-common/hail-init.sh')
        actions.append('gs://dataproc-initialization-actions/conda/bootstrap-conda.sh')
    if jupyter: # now default
        #if not hail01:
        #   actions.append('gs://hail-common/hailctl/dataproc/0.2.16/init_notebook.py')
        #else:
        actions.append('gs://hail-common/cloudtools/init_notebook-0.1-4.py') # actions.append('gs://hail-common/cloudtools/init_notebook1.py')
    if vep:
        if hail01:
            #actions.append('gs://tsingh/bin/vep85-init.sh')
            actions.append('gs://hail-common/vep/vep/vep85-init.sh')
        else:
            actions.append('gs://hail-common/vep/vep/vep{vep_version}-loftee-1.0-{vep_ref}-init-docker.sh'.format(
                vep_version = 85 if vep_reference == 'GRCh37' else 95,
                vep_ref = vep_reference))
    if repos:
        actions.append('gs://tsingh/bin/init_repos.sh')
    if r: 
        actions.append('gs://gnomad-public/tools/inits/master-init.sh')
    return(actions) 

def hail_paths(version = 2):
    hail_version = '0.1' if version == 1 else '0.2'
    spark_version = '2.0.2' if version == 1 else '2.4.0' # goes with image version 1.1

    hail_bucket = 'gs://hail-common/builds/{}'.format(hail_version)    
    hail_hash_path = '{}/latest-hash-spark-{}.txt'.format(hail_bucket, spark_version) if hail_version == '0.1' \
        else '{}/latest-hash/cloudtools-5-spark-{}.txt'.format(hail_bucket, spark_version)
    hail_hash = execute(['gsutil', 'cat', hail_hash_path], return_out = True, silent = True)[0].strip()
    hail_jar = '{0}/jars/hail-{1}-{2}-Spark-{3}.jar'.format(hail_bucket, hail_version, hail_hash, spark_version)
    pyhail_zip = '{0}/python/hail-{1}-{2}.zip'.format(hail_bucket, hail_version, hail_hash)
    
    if version == 1:
        pyhail_zip = 'gs://hail-common/builds/0.1/python/hail-0.1-5a6778710097.zip'
        hail_jar = 'gs://hail-common/builds/0.1/jars/hail-0.1-5a6778710097-Spark-2.0.2.jar'
    return(hail_jar, pyhail_zip)

class DataProcInstance():
    def __init__(self, instance, account, project,               
                 zone = 'us-central1-a',
                 master_machine_type = 'n1-highmem-8', 
                 master_boot_disk_size = 100, 
                 n_workers = 2, 
                 worker_machine_type = 'n1-standard-8', # 'n1-highmem-8' for VEP
                 n_worker_local_ssds = 1,
                 worker_boot_disk_size = 75, # 40
                 n_pre_workers = 2, 
                 version = 2, 
                 properties = {'spark:spark.driver.extraJavaOptions': '-Xss4M',
                               'spark:spark.executor.extraJavaOptions': '-Xss4M',
                               'spark:spark.driver.memory': '45g', 
                               'spark:spark.driver.maxResultSize': '30g',
                               'spark:spark.task.maxFailures': '20',
                               'spark:spark.kryoserializer.buffer.max': '1g',
                               'hdfs:dfs.replication': '1',
                               'dataproc:dataproc.logging.stackdriver.enable': 'false',
                               'dataproc:dataproc.monitoring.stackdriver.enable': 'false'},
                 initialization_actions = [ 'gs://hail-common/hail-init.sh' ],
                 packages = '',
                 hail_jar_path = None, hail_zip_path = None,
                 create = True,
                 dry_run = False):
        
        # Store variables
        self.account = account
        self.instance = instance
        self.project = project
        self.zone = zone
        self.master_machine_type = master_machine_type
        self.master_boot_disk_size = master_boot_disk_size
        self.n_workers = n_workers
        self.worker_machine_type = worker_machine_type
        self.n_worker_local_ssds = n_worker_local_ssds
        self.worker_boot_disk_size = worker_boot_disk_size
        self.n_pre_workers = n_pre_workers
        self.version = version
        self.image_version = '1.1.121-debian9' if version == 1 else '1.4-debian9' # froze the Hail 0.1 # 1.1.73 used to work...
        self.properties = ','.join([ '{}={}'.format(key, properties[key]) for key in properties ])
        self.initialization_actions = ','.join(initialization_actions)

        hail_jar, pyhail_zip = hail_paths(version = version)

        if hail_jar_path and hail_zip_path:
            hail_jar = hail_jar_path
            pyhail_zip = hail_zip_path

        # prepare metadata values
        self.metadata = 'JAR={0},ZIP={1}'.format(hail_jar, pyhail_zip)

        #self.packages = 'numpy<2|pandas>0.22,<0.2 4|bokeh>1.1,<1.3|parsimonious<0.9|ipykernel<5|decorator<5|requests>=2.21.0,<2.21.1|gcsfs==0.2.1|hurry.filesize==0.9|scipy>1.2,<1.4|'
        self.packages = packages # self.packages + 

        if version == 1:
            self.metadata = self.metadata + ',MINICONDA_VARIANT=2'
        else:
            self.metadata = self.metadata + ',MINICONDA_VERSION=4.4.10'
        
        #if self.packages:
        #   self.metadata = self.metadata + '|||PKGS="{}"'.format('|'.join(self.packages.strip('|').split("|")))
            
        # Create dataproc instance]
        cmd = self.create_instance_01() if version == 1 else self.create_instance()

        self.sc = SlackConnector('PyCloud')

        if create:
            print(cmd)
            execute([ str(c) for c in cmd ])
            self.sc.notify('Cluster %s has been created.' % self.instance)
        
    def create_instance(self):
        cmd = [ 
            'hailctl', 'dataproc', 'start', self.instance, 
            '--project', self.project,
            '--zone', self.zone,
            '--master-machine-type', self.master_machine_type,
            '--master-memory-fraction', 0.8,
            '--master-boot-disk-size', self.master_boot_disk_size,
            '--num-master-local-ssds', 0,
            '--num-preemptible-workers', self.n_pre_workers,
            '--num-worker-local-ssds', self.n_worker_local_ssds,
            '--num-workers', self.n_workers,
            '--preemptible-worker-boot-disk-size', 40,
            '--worker-boot-disk-size', self.worker_boot_disk_size, 
            '--worker-machine-type', self.worker_machine_type,
            '--properties', self.properties,
            '--metadata', self.metadata,
            '--init', self.initialization_actions,
        ]
        return(cmd)

    def create_instance_01(self):
        cmd = ['gcloud', 'dataproc', 'clusters','create', 
               self.instance,
               '--project', self.project,
               '--zone', self.zone,
               '--master-machine-type', self.master_machine_type,
               '--master-boot-disk-size', self.master_boot_disk_size,
               '--num-workers', self.n_workers,
               '--worker-machine-type', self.worker_machine_type,
               '--worker-boot-disk-size', self.worker_boot_disk_size, 
               '--num-worker-local-ssds', self.n_worker_local_ssds,
               '--num-preemptible-workers', self.n_pre_workers,
               '--image-version', self.image_version,
               '--initialization-actions', self.initialization_actions,
               '--properties', self.properties,
               '--metadata', self.metadata
              ]      
        return(cmd)

    def describe(self):
        output = execute(['gcloud', 'dataproc', 'clusters', 'describe', self.instance], 
                         silent = True, return_out = True)
        return(yaml.safe_load(''.join(output)))
        
    def diagnose(self):
        execute(['gcloud', 'dataproc', 'clusters', 'diagnose', self.instance])
    
    def refresh(self):
        dp = DataProc(self.account, self.project)
        self.__dict__ = dp.DataProcInstance(self.instance).__dict__
    
    def update(self, n_workers = None, n_pre_workers = None):
        cmd = [ 'gcloud', 'dataproc', 'clusters', 'update', self.instance ] 
        if n_workers:
            cmd.extend(['--num-workers', n_workers])
            self.n_workers = n_workers
        if n_pre_workers:
            cmd.extend(['--num-preemptible-workers', n_pre_workers])
            self.n_pre_workers = n_pre_workers
        execute(cmd)
        self.sc.notify('Cluster %s has been updated.' % self.instance)

    def time_warning(self):
        return
    
    def delete(self):
        execute(' '.join(['yes', '|', 'gcloud', 'dataproc', 'clusters', 'delete', self.instance ]), shell = True)
        self.sc.notify('Cluster %s has been deleted.' % self.instance)
        self.instance = None

        return
    
    def ssh(self):
        cmd = ['gcloud', 'compute', 'ssh',
               '{}-m'.format(self.instance),
               '--zone={}'.format(self.zone),
               '--ssh-flag="-D 10000 -N -f -n"'
               '> /dev/null 2>&1 &']
        print(execute(cmd, return_cmd = True))
    
    def connect(self, port):
        browser_exec = '/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome'
        cmd = [ browser_exec,
                'http://localhost:{}'.format(port),
                '--proxy-server="socks5://localhost:10000"',
                '--host-resolver-rules="MAP * 0.0.0.0 , EXCLUDE localhost"',
                '--user-data-dir=/tmp/',
                '> /dev/null 2>&1 &' ]
        
        print(execute(cmd, return_cmd = True))
    
    def notebook(self):
        self.connect(8123)
        
    def sparkui(self):
        self.connect(4040)
    
    def list_jobs(self, return_out = False):
        cmd = [ 'gcloud', 'dataproc', 'jobs', 'list']
        cmd.append("--cluster={}".format(self.instance))
        cmd.append("--state-filter=active")
        if not return_out:
            execute(cmd)
        else:
            output = execute(cmd, return_out = return_out, silent = True)
            return(output)
        
    def number_of_active_jobs(self):
        return(len([ 0 for line in self.list_jobs(return_out = True) if re.search('RUNNING', line) ]))